from enum import Enum


class JobStatus(Enum):
    """JobStatus enum."""
    READY = "READY"
    PENDING = "PENDING"
    RUNNING = "RUNNING"
    FINISHED = "FINISHED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"



from pydantic import BaseModel


class BaseRequest(BaseModel):
    """Base class for API requests."""

    @classmethod
    def example(cls):
        empty_cls = cls()
        field_values = empty_cls.model_dump()
        for name, field in cls.model_fields.items():
            default_value = field.default
            field_values[name] = default_value
            if field.examples is not None:
                if len(field.examples) > 0:
                    field_values[name] = field.examples[0]
        return cls(**field_values)



import asyncio
import base64
import dataclasses
import json
import logging
import os
import urllib
import urllib.parse
import urllib.request
from typing import Any, Dict, List, Optional, Union

import httpx
import requests

from creator.base.base_app import BaseApp
from creator.base.base_request import BaseRequest
from creator.base.base_response import ResponseDataType, ResponseData, BaseResponse
from creator.base.job_status import JobStatus
from creator.cmfy.request_cmfy import CmfyWorkflow, CmfyWorkflowFcus
from creator.cmfy.response_cmfy import ResponseCmfy

logger = logging.getLogger(__name__)


class AppCmfy(BaseApp):
    param_classes = [CmfyWorkflow, CmfyWorkflowFcus]
    output = {}

    def __init__(self):
        super().__init__()
        focus_port = os.environ.get("PORT_CMFY", 8888)
        if isinstance(focus_port, str):
            focus_port = int(focus_port)
        self.api_base_url = f"http://0.0.0.0:{focus_port}"

    async def create(self, params: Union[CmfyWorkflow, CmfyWorkflowFcus]) -> ResponseCmfy:
        cmfy_port = os.environ.get("PORT_CMFY", 8889)
        url = f"http://localhost:{cmfy_port}/prompt"
        workflow = json.loads(params.workflow_json)
        self._check_loaded_model(workflow)
        logger.debug(f"CMFY data: {workflow}")
        # Send a POST to cmfy_url with the workflow json
        # The response will be JSON string with the below format
        try:
            headers = {'Content-Type': 'application/json'}
            data = {"prompt": workflow}
            data = json.dumps(data).encode('utf-8')
            response = requests.post(url, data=data, headers=headers)
            logger.debug(f"CMFY response: {response}, {response.text}, {response.json()}")
            # If we have a 422 here, print what the unprocessable entity is
            if response.status_code == 422:
                logger.warning(f"Unprocessable entity: {response.text}")
            response.raise_for_status()
            response_json = response.json()
        except Exception as e:
            logger.error(f"Error creating prompt: {e}")
            return ResponseCmfy.error("Error creating prompt")
        # {"prompt_id": "f57c69ee-f4ed-4866-b5a4-4072c17c36a8", "number": 2, "node_errors": {}}
        prompt_id = response_json.get("prompt_id", None)
        logger.debug(f"Prompt ID: {prompt_id}")
        if prompt_id is None:
            return ResponseCmfy.error("Error creating prompt")
        status_count = 0
        status = await self.get_status(prompt_id)
        while status != JobStatus.FINISHED and status != JobStatus.FAILED and status_count < 10:
            await asyncio.sleep(1)
            status = await self.get_status(prompt_id)
            status_count += 1
        if status == JobStatus.FINISHED:
            data = self.output
            response_data = ResponseData(
                data=data,
                data_type=ResponseDataType.IMAGE,
                total_count=len(data)
            )
            logger.info(f"Success, returning response with {len(data)} images")
            return ResponseCmfy.success(response_data, prompt_id)
        return ResponseCmfy.error("Error creating prompt")

    async def create_async(self, params: Union[CmfyWorkflow, CmfyWorkflowFcus]) -> ResponseCmfy:
        cmfy_port = os.environ.get("PORT_CMFY", 8889)
        url = f"http://localhost:{cmfy_port}/prompt"
        workflow = json.loads(params.workflow_json)
        self._check_loaded_model(workflow)
        logger.debug(f"CMFY data: {workflow}")
        # Send a POST to cmfy_url with the workflow json
        # The response will be JSON string with the below format
        try:
            headers = {'Content-Type': 'application/json'}
            data = {"prompt": workflow}
            data = json.dumps(data).encode('utf-8')
            response = requests.post(url, data=data, headers=headers)
            logger.debug(f"CMFY response: {response}, {response.text}, {response.json()}")
            # If we have a 422 here, print what the unprocessable entity is
            if response.status_code == 422:
                logger.warning(f"Unprocessable entity: {response.text}")
            response.raise_for_status()
            response_json = response.json()
        except Exception as e:
            logger.error(f"Error creating prompt: {e}")
            return ResponseCmfy.error("Error creating prompt")
        # {"prompt_id": "f57c69ee-f4ed-4866-b5a4-4072c17c36a8", "number": 2, "node_errors": {}}
        prompt_id = response_json.get("prompt_id", None)
        logger.debug(f"Prompt ID: {prompt_id}")
        if prompt_id is None:
            return ResponseCmfy.error("Error creating prompt")
        return ResponseCmfy.running(prompt_id)

    async def get_status(self, job_id) -> BaseResponse:
        url = f"http://localhost:{os.environ.get('PORT_CMFY', 8889)}/history/{job_id}"
        job_data = None
        print(f"Getting status for job {job_id} from url {url}")
        logger.debug(f"Getting status for job {job_id}")
        while job_data is None:
            history = requests.get(url)
            if history.status_code != 200:
                return BaseResponse.error("Error getting job status", job_id)
            logger.info(f"History: {history.json()}")
            history = history.json()
            for history_id, job in history.items():
                if history_id == job_id:
                    job_data = job
                    break
        images_output = []
        for node_id, node_output in job_data['outputs'].items():
            if 'images' in node_output:
                logger.debug(f"Node output: {node_output}")
                for image in node_output['images']:
                    logger.debug(f"Getting image: {image['filename']}")
                    print(f"Getting image: {image['filename']}")
                    image_data = await self._get_image(image['filename'], image['subfolder'], image['type'])
                    images_output.append(image_data)
                print(f"We are returning {len(images_output)} images")
                response_data = ResponseData(data=images_output, data_type=ResponseDataType.IMAGE,
                                             total_count=len(images_output))
                return ResponseCmfy.success(response_data, job_id)
        response = ResponseCmfy.running(job_id)
        return response

    async def get_models(self) -> List[Any]:
        url = f"http://localhost:{os.environ.get('PORT_CMFY', 8889)}/object_info"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            response.raise_for_status()
            response_json = response.json()
            models = []
            loader = response_json.get("CheckpointLoaderSimple", None)
            if loader:
                input = loader.get("input", None)
                if input:
                    required = input.get("required", None)
                    if required:
                        model_name = required.get("ckpt_name", None)
                        if model_name:
                            models.append(model_name)
        return models

    async def upload_image(self, image_data: Any) -> str:
        pass

    async def select_model(self, params: BaseRequest):
        pass

    async def test(self):
        url = f"http://localhost:{os.environ.get('PORT_CMFY', 8889)}/history"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            if response.status_code == 200:
                return BaseResponse.active()
            else:
                return BaseResponse.error("Error testing connection")

    @staticmethod
    async def _get_image(filename: str, subfolder: str, folder_type: str) -> Optional[str]:
        server_address = f"localhost:{os.environ.get('PORT_CMFY', 8889)}"
        data = {"filename": filename, "subfolder": subfolder, "type": folder_type}
        url_values = urllib.parse.urlencode(data)
        req_url = f"http://{server_address}/view?{url_values}"
        logger.debug(f"Getting image from {req_url}")
        try:
            async with httpx.AsyncClient() as client:
                response = await client.get(req_url)
                response.raise_for_status()
                image_data = response.content
                # Convert the image data to a base64 string
                base64_encoded = base64.b64encode(image_data).decode('utf-8')
                print(f"Image data: {base64_encoded[:100]}")
                return base64_encoded
        except Exception as e:
            logger.error(f"Failed to retrieve or encode image: {e}")
            return None

    def _check_loaded_model(self, workflow: Dict[str, Any]):
        for node_id, node in workflow.items():
            try:
                if node['class_type'] == "CheckpointLoaderSimple":
                    model_name = node['inputs']['ckpt_name']
                    if model_name and model_name not in self.loaded_models:
                        self.loaded_models.append(model_name)
            except KeyError:
                continue
        max_cached_models = os.environ.get("MAX_CACHED_MODELS", 3)
        if len(self.loaded_models) > max_cached_models:
            self.loaded_models = self.loaded_models[-max_cached_models:]



from typing import Dict, Any
from creator.base.base_response import BaseResponse, ResponseData, ResponseDataType


class ResponseCmfy(BaseResponse):
    @classmethod
    def parse_response(cls, response: Dict[str, Any]):
        data = response.get('images', [])
        if not data:
            return cls.error("No images found in response")
        response_data = ResponseData(data=data, data_type=ResponseDataType.IMAGE, total_count=len(data))
        return cls.success(response_data)



from typing import List, Dict, Any

from pydantic import Field

from creator.base.base_request import BaseRequest


class AutoTxt2Img(BaseRequest):
    prompt: str = Field("", examples=["A beautiful sunset over a mountain lake."])
    negative_prompt: str = ""
    styles: List[str] = Field(default_factory=list, examples=[[]])
    seed: int = -1
    subseed: int = -1
    subseed_strength: float = 0.0
    seed_resize_from_h: int = -1
    seed_resize_from_w: int = -1
    sampler_name: str = "Euler a"
    batch_size: int = 1
    n_iter: int = 1
    steps: int = 20
    cfg_scale: float = 7.0
    width: int = 1024
    height: int = 768
    restore_faces: bool = False
    tiling: bool = False
    do_not_save_samples: bool = True
    do_not_save_grid: bool = True
    eta: float = 0.0
    denoising_strength: float = 0.7
    s_min_uncond: float = 0.0
    s_churn: float = 0.0
    s_tmax: float = 0.0
    s_tmin: float = 0.0
    s_noise: float = 1.0
    override_settings: Dict[str, Any] = Field(default_factory=dict, examples=[{}])
    override_settings_restore_afterwards: bool = True
    refiner_checkpoint: str = ""
    refiner_switch_at: int = 0
    disable_extra_networks: bool = False
    firstpass_image: str = ""
    comments: Dict[str, Any] = Field(default_factory=dict, examples=[{}])
    enable_hr: bool = False
    firstphase_width: int = 0
    firstphase_height: int = 0
    hr_scale: int = 2
    hr_upscaler: str = "Latent"
    hr_second_pass_steps: int = 0
    hr_resize_x: int = 0
    hr_resize_y: int = 0
    hr_checkpoint_name: str = ""
    hr_sampler_name: str = ""
    hr_prompt: str = ""
    hr_negative_prompt: str = ""
    force_task_id: str = ""
    sampler_index: str = "0"
    script_name: str = ""
    script_args: List[Any] = Field(default_factory=list, examples=[[]])
    send_images: bool = True
    save_images: bool = False
    alwayson_scripts: Dict[str, Any] = Field(default_factory=dict, examples=[{}])
    infotext: str = ""


class AutoImg2Img(BaseRequest):
    prompt: str = ""
    negative_prompt: str = ""
    styles: List[str] = Field(default_factory=list, examples=[[]])
    seed: int = -1
    subseed: int = -1
    subseed_strength: float = 0.0
    seed_resize_from_h: int = -1
    seed_resize_from_w: int = -1
    sampler_name: str = ""
    scheduler: str = ""
    batch_size: int = 1
    n_iter: int = 1
    steps: int = 50
    cfg_scale: float = 7.0
    width: int = 512
    height: int = 512
    restore_faces: bool = True
    tiling: bool = True
    do_not_save_samples: bool = False
    do_not_save_grid: bool = False
    eta: float = 0.0
    denoising_strength: float = 0.75
    s_min_uncond: float = 0.0
    s_churn: float = 0.0
    s_tmax: float = 0.0
    s_tmin: float = 0.0
    s_noise: float = 0.0
    override_settings: Dict[str, Any] = Field(default_factory=dict, examples=[{}])
    override_settings_restore_afterwards: bool = True
    refiner_checkpoint: str = ""
    refiner_switch_at: int = 0
    disable_extra_networks: bool = False
    firstpass_image: str = ""
    comments: Dict[str, Any] = Field(default_factory=dict, examples=[{}])
    init_images: List[str] = Field(default_factory=list, examples=[[]])
    resize_mode: int = 0
    image_cfg_scale: float = 0.0
    mask: str = ""
    mask_blur_x: int = 4
    mask_blur_y: int = 4
    mask_blur: int = 0
    mask_round: bool = True
    inpainting_fill: int = 0
    inpaint_full_res: bool = True
    inpaint_full_res_padding: int = 0
    inpainting_mask_invert: int = 0
    initial_noise_multiplier: float = 0.0
    latent_mask: str = ""
    force_task_id: str = ""
    sampler_index: str = "Euler"
    include_init_images: bool = False
    script_name: str = ""
    script_args: List[Any] = Field(default_factory=list, examples=[[]])
    send_images: bool = True
    save_images: bool = False
    alwayson_scripts: Dict[str, Any] = Field(default_factory=dict, examples=[{}])
    infotext: str = ""


class AutoSchedTxt2Img(BaseRequest):
    prompt: str = Field("", examples=["A beautiful sunset over a mountain lake."])
    negative_prompt: str = ""
    styles: List[str] = Field(default_factory=list, examples=[[]])
    seed: int = -1
    subseed: int = -1
    subseed_strength: int = 0
    seed_resize_from_h: int = -1
    seed_resize_from_w: int = -1
    sampler_name: str = "DPM++ 2M"
    scheduler: str = ""
    batch_size: int = 1
    n_iter: int = 1
    steps: int = 50
    cfg_scale: int = 7
    width: int = 1024
    height: int = 768
    restore_faces: bool = True
    tiling: bool = True
    do_not_save_samples: bool = False
    do_not_save_grid: bool = False
    eta: int = 0
    denoising_strength: int = 0
    s_min_uncond: int = 0
    s_churn: int = 0
    s_tmax: int = 0
    s_tmin: int = 0
    s_noise: int = 0
    override_settings: Dict[str, Any] = Field(default_factory=dict, examples=[{}])
    override_settings_restore_afterwards: bool = True
    refiner_checkpoint: str = ""
    refiner_switch_at: int = 0
    disable_extra_networks: bool = False
    firstpass_image: str = ""
    comments: Dict[str, Any] = {}
    enable_hr: bool = False
    firstphase_width: int = 0
    firstphase_height: int = 0
    hr_scale: int = 2
    hr_upscaler: str = ""
    hr_second_pass_steps: int = 0
    hr_resize_x: int = 0
    hr_resize_y: int = 0
    hr_checkpoint_name: str = ""
    hr_sampler_name: str = ""
    hr_scheduler: str = ""
    hr_prompt: str = ""
    hr_negative_prompt: str = ""
    force_task_id: str = ""
    script_name: str = ""
    script_args: List[str] = []
    alwayson_scripts: Dict[str, Any] = {}
    infotext: str = ""
    checkpoint: str = "sdxl/juggernautXL_v9Rundiffusionphoto2.safetensors"
    vae: str = ""
    callback_url: str = ""


class AutoModelLoad(BaseRequest):
    sd_model_checkpoint: str = Field("", examples=["sdxl/juggernautXL_v9Rundiffusionphoto2.safetensors"])



import asyncio
import logging
import os
import random
import traceback
from typing import Any, List, Union

import httpx

from creator.auto.request_auto import AutoTxt2Img, AutoModelLoad, AutoImg2Img, AutoSchedTxt2Img
from creator.auto.response_auto import ResponseAuto
from creator.base.base_app import BaseApp
from creator.base.base_response import BaseResponse

logger = logging.getLogger(__name__)


class AppAuto(BaseApp):
    param_classes = [AutoTxt2Img, AutoImg2Img, AutoSchedTxt2Img]
    output = {}
    jobs = {}
    sched_jobs = {}

    def __init__(self):
        super().__init__()
        auto_port = os.environ.get("PORT_AUTO", 8081)
        if isinstance(auto_port, str):
            auto_port = int(auto_port)
        self.jobs = {}
        self.output = {}
        self.api_base_url = f"http://0.0.0.0:{auto_port}/sdapi"

    async def create(self, params: Union[AutoTxt2Img, AutoImg2Img, AutoSchedTxt2Img], job_id=None) -> BaseResponse:
        print(f"Creating job with params: {params}")
        endpoint_name = "txt2img" if isinstance(params, AutoTxt2Img) else "img2img"
        url = f"{self.api_base_url}/sdapi/v1/{endpoint_name}"
        get_task_id = False
        if isinstance(params, AutoSchedTxt2Img):
            url = f"{self.api_base_url}/agent-scheduler/v1/queue/tx2img"
            get_task_id = True
        data = params.__dict__
        headers = {'Content-Type': 'application/json'}
        async with httpx.AsyncClient() as client:
            response = await client.post(url, data=data, headers=headers)
            response.raise_for_status()
        output = response.json()
        print(f"Response received: {output}")
        if job_id:
            self.output[job_id] = output
        self.jobs[job_id] = output
        return ResponseAuto.parse_response(output, job_id)

    async def create_async(self, params: Union[AutoTxt2Img, AutoImg2Img, AutoSchedTxt2Img]) -> BaseResponse:
        job_id = str(random.randint(0, 100000))
        print(f"Creating job with ID {job_id}")

        # Store the job_id immediately with a placeholder to indicate it's pending
        self.output[job_id] = {"status": "pending"}

        # Fire off the job without waiting for it to complete
        asyncio.create_task(self._create_job(params, job_id, self.output))

        print(f"Job {job_id} is running.")
        return ResponseAuto.running(job_id)

    async def get_status(self, job_id: str) -> BaseResponse:
        try:
            output = self.output.get(job_id)
            if output:
                task_id = output.get("task_id")
                if task_id:
                    url = f"{self.api_base_url}/agent-scheduler/v1/queue/tx2img/{task_id}"
                    async with httpx.AsyncClient() as client:
                        response = await client.get(url)
                        response.raise_for_status()
                    output = response.json()
                    return ResponseAuto.parse_response(output, job_id)
            else:
                # Check if the job resulted in an error
                if "error" in output:
                    print(f"Error found for job ID {job_id}")
                    return ResponseAuto.error(output["error"], job_id)
                print(f"Output found for job ID {job_id}")
                return ResponseAuto.parse_response(output, job_id)

            # If job_id is not in the jobs dictionary and no output is found, assume the job does not exist
            if job_id not in self.jobs:
                logger.warning(f"Job ID {job_id} not found")
                return ResponseAuto.error(f"Job ID {job_id} not found", job_id)

            # If the job is still in jobs but no output has been generated yet
            print(f"Job {job_id} is still running.")
            return ResponseAuto.running(job_id)

        except Exception as e:
            logger.error(f"Error getting job status for job ID {job_id}: {e}")
            traceback.print_exc()
            return ResponseAuto.error(f"Error getting job status: {e}", job_id)

    async def get_models(self) -> List[Any]:
        url = f"{self.api_base_url}/sdapi/v1/sd-models"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            response.raise_for_status()
            response_json = response.json()
        models = [model.get("title") for model in response_json if model.get("title")]
        return models

    # Assuming you will provide more detail or use this method later
    async def upload_image(self, image_data: Any) -> str:
        pass

    async def select_model(self, params: AutoModelLoad):
        model_to_load = params.sd_model_checkpoint
        if model_to_load and model_to_load not in self.loaded_models:
            self.loaded_models.append(model_to_load)
        max_cached_models = os.environ.get("MAX_CACHED_MODELS", 3)
        if len(self.loaded_models) > max_cached_models:
            self.loaded_models = self.loaded_models[-max_cached_models:]
        url = f"{self.api_base_url}/sdapi/v1/options"
        headers = {'Content-Type': 'application/json'}
        async with httpx.AsyncClient() as client:
            response = await client.post(url, data=params.__dict__, headers=headers)
            response.raise_for_status()
            return response.json()

    async def test(self):
        auto_port = int(os.environ.get("PORT_AUTO", 8081))
        url = f"http://0.0.0.0:{auto_port}/docs"
        async with httpx.AsyncClient() as client:
            response = await client.get(url)
            if response.status_code == 200:
                return BaseResponse.active()
            else:
                return BaseResponse.error("Error testing connection")

    async def _create_job(self, params: AutoTxt2Img, job_id: str, output_store: dict):
        url = f"{self.api_base_url}/v1/txt2img"  # Adjust URL as necessary
        data = params.__dict__
        headers = {'Content-Type': 'application/json'}
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(url, data=data, headers=headers)
                response.raise_for_status()
            output = response.json()
            print(f"Response received: {output}")
            output_store[job_id] = output  # Storing output for future retrieval
        except Exception as e:
            print(f"Error in job {job_id}: {e}")
            output_store[job_id] = {"error": str(e)}



from abc import abstractmethod, ABC
from typing import Any, List

from creator.base.base_request import BaseRequest
from creator.base.base_response import BaseResponse, ResponseData, ResponseDataType
from creator.base.job_status import JobStatus


class BaseApp(ABC):
    def __init__(self):
        self.output = None
        self.processing = False
        self.can_queue = False
        self.loaded_models = []

    @abstractmethod
    async def create(self, params: BaseRequest) -> BaseResponse:
        """Create a job and await its completion."""
        pass

    @abstractmethod
    async def create_async(self, params: BaseRequest) -> BaseResponse:
        """Create an asynchronous job and return the job_id."""
        pass

    @abstractmethod
    async def get_status(self, job_id: str) -> BaseResponse:
        """Get the status of an asynchronous job."""
        pass

    @abstractmethod
    async def get_models(self) -> List[Any]:
        pass

    async def get_loaded_models(self) -> BaseResponse:
        base_response = BaseResponse(status=JobStatus.FINISHED)
        base_response_data = ResponseData(data=self.loaded_models, data_type=ResponseDataType.TEXT, total_count=len(self.loaded_models))
        base_response.output = base_response_data
        return base_response

    @abstractmethod
    def upload_image(self, image_data: Any) -> str:
        pass

    @abstractmethod
    async def select_model(self, params: BaseRequest):
        pass

    @abstractmethod
    async def test(self):
        pass



import importlib
import pkgutil
import sys
from typing import Union, Type, Dict, List

from creator.base.base_app import BaseApp
from creator.base.base_request import BaseRequest


def get_creators():
    package = sys.modules['creator']  # Make sure 'src.sb_api.creator' is the correct package path
    creator_names = []
    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        # Filter out unwanted modules
        if "creator_" not in modname or "base" in modname:
            continue

        # Try to import the module
        try:
            module = importlib.import_module(modname)
            # Find all subclasses of BaseApp in the imported module
            for name, obj in vars(module).items():
                if isinstance(obj, type) and issubclass(obj, BaseApp) and obj is not BaseApp:
                    base_app_name = obj.__name__.replace("App", "")
                    base_app_name = ''.join(['_' + i.lower() if i.isupper() else i for i in base_app_name]).lstrip('_')
                    creator_names.append(base_app_name)
        except Exception as e:
            print(f"Error importing {modname}: {e}")
    return creator_names


def get_creator_examples(openapi_output=False):
    creators = get_creators()
    examples = []
    openapi_examples = {}
    for creator in creators:
        examples.append({"name": creator, "description": f"Example for {creator} creator."})
        openapi_examples[creator] = {
            "summary": f"{creator}",
            "value": creator
        }
    return examples if not openapi_output else openapi_examples


def get_use_cases(return_creator_names: bool = False, return_param_names: bool = False) -> Union[Dict[str, List[BaseRequest]], Dict[BaseApp, List[BaseRequest]]]:
    package = sys.modules['creator']  # Make sure 'src.sb_api.creator' is the correct package path
    use_case_dict = {}

    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        # Filter out unwanted modules
        if "creator_" not in modname or "base" in modname:
            continue

        # Try to import the module
        try:
            module = importlib.import_module(modname)
            # Find all subclasses of BaseApp in the imported module
            for name, obj in vars(module).items():
                if isinstance(obj, type) and issubclass(obj, BaseApp) and obj is not BaseApp:
                    params = getattr(obj, "param_classes", [])
                    if return_param_names:
                        params = [param.__name__ for param in params]
                    if not return_creator_names:
                        use_case_dict[obj] = params
                    else:
                        use_case_dict[obj.__name__] = params
        except Exception as e:
            print(f"Error importing {modname}: {e}")

    return use_case_dict


def get_use_case_examples(return_class_name=False, openapi_output=False):
    use_cases = get_use_cases(True, False)  # Always fetch class objects
    examples = []
    openapi_examples = {}
    all_use_cases = []
    for creator, param_classes in use_cases.items():
        all_use_cases.extend(param_classes)
    for param_class in all_use_cases:
        example_param = param_class().example()  # Instantiate the class object
        example_execution_params = {
            "params": example_param,
            "callback_url": "http://example.com/callback"
        }
        summary = f"Execute {param_class.__name__} with optional callback"
        example_name = param_class.__name__ + ("_async" if return_class_name else "")
        examples.append({"name": example_name, "value": example_execution_params})
        openapi_examples[example_name] = {
            "summary": summary,
            "value": example_execution_params
        }
    return examples if not openapi_output else openapi_examples


def get_method_examples(openapi_output=False):
    use_cases = get_use_cases(True, True)
    examples = []
    openapi_examples = {}
    all_use_cases = []
    for creator, param_classes in use_cases.items():
        all_use_cases.extend(param_classes)
    for method in all_use_cases:
        examples.append({"name": method, "description": f"Example for {method} method."})
        openapi_examples[method] = {
            "summary": f"{method}",
            "value": method
        }
    return examples if not openapi_output else openapi_examples


def get_use_case(app_name: str, method: str, debug=False) -> Union[None, Type[BaseRequest]]:
    package = sys.modules['creator']  # Make sure 'src.sb_api.creator' is the correct package path

    for importer, modname, ispkg in pkgutil.walk_packages(package.__path__, package.__name__ + '.'):
        # Filter out unwanted modules
        if "creator_" not in modname or "base" in modname:
            continue

        # Try to import the module
        try:
            module = importlib.import_module(modname)
            # Find all subclasses of BaseApp in the imported module
            for name, obj in vars(module).items():
                if isinstance(obj, type) and issubclass(obj, BaseApp) and obj is not BaseApp:
                    base_app_name = obj.__name__.replace("App", "")
                    base_app_name = ''.join(['_' + i.lower() if i.isupper() else i for i in base_app_name]).lstrip('_')
                    if base_app_name != app_name:
                        continue
                    params = getattr(obj, "param_classes", [])
                    for param in params:
                        # Compare the method name to the base name of the param class
                        param_base_name = param.__name__.split("_")[-1]
                        if param_base_name == method:
                            return param
        except Exception as e:
            print(f"Error importing {modname}: {e}")
    return None



import json

from pydantic import Field

from creator.base.base_request import BaseRequest

default_workflow_json = {
    "3": {
        "inputs": {
            "seed": 985289875893903,
            "steps": 20,
            "cfg": 7.5,
            "sampler_name": "dpmpp_2m_sde",
            "scheduler": "karras",
            "denoise": 1,
            "model": [
                "4",
                0
            ],
            "positive": [
                "6",
                0
            ],
            "negative": [
                "7",
                0
            ],
            "latent_image": [
                "5",
                0
            ]
        },
        "class_type": "KSampler",
        "_meta": {
            "title": "KSampler"
        }
    },
    "4": {
        "inputs": {
            "ckpt_name": "Juggernaut-XI-Prototype.safetensors"
        },
        "class_type": "CheckpointLoaderSimple",
        "_meta": {
            "title": "Load Checkpoint"
        }
    },
    "5": {
        "inputs": {
            "width": 1024,
            "height": 768,
            "batch_size": 1
        },
        "class_type": "EmptyLatentImage",
        "_meta": {
            "title": "Empty Latent Image"
        }
    },
    "6": {
        "inputs": {
            "text": "a photograph of a smiling green alien holding a sign which says \"Run Diffusion\", ultra realistic, street, city, sci-fi, ultra realistic, cute, friendly",
            "clip": [
                "4",
                1
            ]
        },
        "class_type": "CLIPTextEncode",
        "_meta": {
            "title": "CLIP Text Encode (Prompt)"
        }
    },
    "7": {
        "inputs": {
            "text": "blurry, low quality",
            "clip": [
                "4",
                1
            ]
        },
        "class_type": "CLIPTextEncode",
        "_meta": {
            "title": "CLIP Text Encode (Prompt)"
        }
    },
    "8": {
        "inputs": {
            "samples": [
                "3",
                0
            ],
            "vae": [
                "4",
                2
            ]
        },
        "class_type": "VAEDecode",
        "_meta": {
            "title": "VAE Decode"
        }
    },
    "9": {
        "inputs": {
            "filename_prefix": "ComfyUI",
            "images": [
                "8",
                0
            ]
        },
        "class_type": "SaveImage",
        "_meta": {
            "title": "Save Image"
        }
    }
}

default_workflow_json_2 = {
  "4": {
    "inputs": {
      "ckpt_name": "Juggernaut-XI-Prototype.safetensors"
    },
    "class_type": "CheckpointLoaderSimple",
    "_meta": {
      "title": "Load Checkpoint"
    }
  },
  "21": {
    "inputs": {
      "width": 2432,
      "height": 1664,
      "batch_size": 2
    },
    "class_type": "EmptyLatentImage",
    "_meta": {
      "title": "Image Size\n"
    }
  },
  "67": {
    "inputs": {
      "samples": [
        "132",
        0
      ],
      "vae": [
        "4",
        2
      ]
    },
    "class_type": "VAEDecode",
    "_meta": {
      "title": "VAE Decode"
    }
  },
  "69": {
    "inputs": {
      "images": [
        "67",
        0
      ]
    },
    "class_type": "PreviewImage",
    "_meta": {
      "title": "Output"
    }
  },
  "99": {
    "inputs": {
      "seed": 1042340763738557
    },
    "class_type": "CR Seed",
    "_meta": {
      "title": "🌱 CR Seed"
    }
  },
  "100": {
    "inputs": {
      "text": [
        "114:0",
        0
      ],
      "seed": 1116260329,
      "log_prompt": "No"
    },
    "class_type": "PromptExpansion",
    "_meta": {
      "title": "Prompt Expansion"
    }
  },
  "101": {
    "inputs": {
      "text": [
        "100",
        0
      ]
    },
    "class_type": "ShowText|pysssss",
    "_meta": {
      "title": "Final Positive Prompt"
    }
  },
  "102": {
    "inputs": {
      "text": [
        "114:1",
        0
      ]
    },
    "class_type": "ShowText|pysssss",
    "_meta": {
      "title": "Final Negative Prompt"
    }
  },
  "120": {
    "inputs": {
      "b1": 1.01,
      "b2": 1.02,
      "s1": 0.99,
      "s2": 0.9500000000000001,
      "model": [
        "4",
        0
      ]
    },
    "class_type": "FreeU_V2",
    "_meta": {
      "title": "FreeU_V2"
    }
  },
  "132": {
    "inputs": {
      "add_noise": True,
      "noise_seed": [
        "99",
        0
      ],
      "steps": 20,
      "cfg": 5,
      "sampler_name": "dpmpp_2m",
      "scheduler": "AYS SDXL",
      "start_at_step": 0,
      "end_at_step": 10000,
      "noise_mode": "GPU(=A1111)",
      "return_with_leftover_noise": False,
      "batch_seed_mode": "incremental",
      "variation_seed": 0,
      "variation_strength": 0,
      "model": [
        "134",
        0
      ],
      "positive": [
        "98:0",
        0
      ],
      "negative": [
        "98:1",
        0
      ],
      "latent_image": [
        "21",
        0
      ]
    },
    "class_type": "KSamplerAdvanced //Inspire",
    "_meta": {
      "title": "KSamplerAdvanced (inspire)"
    }
  },
  "134": {
    "inputs": {
      "ds_depth_1": 3,
      "ds_depth_2": 3,
      "ds_timestep_1": 900,
      "ds_timestep_2": 650,
      "resize_scale_1": 2,
      "resize_scale_2": 2,
      "model": [
        "120",
        0
      ]
    },
    "class_type": "Hires",
    "_meta": {
      "title": "Apply Kohya's HiresFix"
    }
  },
  "119:0": {
    "inputs": {
      "text_positive": "",
      "text_negative": "",
      "style": "Fooocus Photograph",
      "log_prompt": True,
      "style_positive": True,
      "style_negative": True
    },
    "class_type": "SDXLPromptStyler",
    "_meta": {
      "title": "SDXL Prompt Styler"
    }
  },
  "119:1": {
    "inputs": {
      "text_positive": "",
      "text_negative": "",
      "style": "Fooocus Enhance",
      "log_prompt": True,
      "style_positive": True,
      "style_negative": True
    },
    "class_type": "SDXLPromptStyler",
    "_meta": {
      "title": "SDXL Prompt Styler"
    }
  },
  "119:2": {
    "inputs": {
      "text_positive": "",
      "text_negative": "",
      "style": "Fooocus Negative",
      "log_prompt": True,
      "style_positive": True,
      "style_negative": True
    },
    "class_type": "SDXLPromptStyler",
    "_meta": {
      "title": "SDXL Prompt Styler"
    }
  },
  "133:0": {
    "inputs": {
      "text1": "a close-up shot of a soldier running on the battlefield, world war 2, gritty, dust, explosions",
      "text2": [
        "119:0",
        0
      ],
      "separator": ", "
    },
    "class_type": "CR Text Concatenate",
    "_meta": {
      "title": "Concat Positive"
    }
  },
  "133:1": {
    "inputs": {
      "text1": "blurry eyes, zombie eyes, bad face, too many fingers",
      "text2": [
        "119:0",
        1
      ],
      "separator": ", "
    },
    "class_type": "CR Text Concatenate",
    "_meta": {
      "title": "Concat Negative"
    }
  },
  "133:2": {
    "inputs": {
      "text1": [
        "119:1",
        0
      ],
      "text2": [
        "119:2",
        0
      ],
      "separator": ", "
    },
    "class_type": "CR Text Concatenate",
    "_meta": {
      "title": "Concat Positive"
    }
  },
  "133:3": {
    "inputs": {
      "text1": [
        "119:1",
        1
      ],
      "text2": [
        "119:2",
        1
      ],
      "separator": ", "
    },
    "class_type": "CR Text Concatenate",
    "_meta": {
      "title": "Concat Negative"
    }
  },
  "114:0": {
    "inputs": {
      "text1": [
        "133:0",
        0
      ],
      "text2": [
        "133:2",
        0
      ],
      "separator": ", "
    },
    "class_type": "CR Text Concatenate",
    "_meta": {
      "title": "Concat Positive"
    }
  },
  "114:1": {
    "inputs": {
      "text1": [
        "133:1",
        0
      ],
      "text2": [
        "133:3",
        0
      ],
      "separator": ", "
    },
    "class_type": "CR Text Concatenate",
    "_meta": {
      "title": "Concat Negative"
    }
  },
  "98:0": {
    "inputs": {
      "text": [
        "101",
        0
      ],
      "clip": [
        "4",
        1
      ]
    },
    "class_type": "CLIPTextEncode",
    "_meta": {
      "title": "Encode Positive"
    }
  },
  "98:1": {
    "inputs": {
      "text": [
        "102",
        0
      ],
      "clip": [
        "4",
        1
      ]
    },
    "class_type": "CLIPTextEncode",
    "_meta": {
      "title": "Encode Negative"
    }
  }
}


class CmfyWorkflow(BaseRequest):
    workflow_json: str = Field("", examples=[json.dumps(default_workflow_json)])


class CmfyWorkflowFcus(BaseRequest):
    workflow_json: str = Field("", examples=[json.dumps(default_workflow_json_2)])
